@@ -1,7 +1,10 @@
 import asyncio
 import dataclasses
+import enum
+import inspect
 import io
 import keyword
+import re
 import sys
 import textwrap
 from dataclasses import MISSING, Field, dataclass
@@ -9,6 +12,7 @@ from io import StringIO
 from typing import Any, Dict, List, Optional, TextIO, Tuple, TypeVar
 
 from strong_typing.auxiliary import python_type_to_str
+from strong_typing.inspection import is_type_enum, is_dataclass_type
 
 from . import async_database
 from .async_database import ConnectionParameters, DatabaseClient
@@ -81,35 +85,41 @@ class _CatalogSchemaBuilder:
     async def get_catalog_schema(self) -> CatalogSchema:
         "Retrieves metadata for the current catalog."
 
+        query = """
+            CREATE TEMPORARY TABLE key_reference AS (
+                SELECT
+                    rcon.unique_constraint_catalog AS primary_constraint_catalog,
+                    rcon.unique_constraint_schema AS primary_constraint_schema,
+                    rcon.unique_constraint_name AS primary_constraint_name,
+                    pkey.table_catalog AS primary_table_catalog,
+                    pkey.table_schema AS primary_table_schema,
+                    pkey.table_name AS primary_table_name,
+                    pkey.column_name AS primary_column_name,
+                    rcon.constraint_catalog AS foreign_constraint_catalog,
+                    rcon.constraint_schema AS foreign_constraint_schema,
+                    rcon.constraint_name AS foreign_constraint_name,
+                    fkey.table_catalog AS foreign_table_catalog,
+                    fkey.table_schema AS foreign_table_schema,
+                    fkey.table_name AS foreign_table_name,
+                    fkey.column_name AS foreign_column_name
+
+                FROM
+                    information_schema.referential_constraints AS rcon
+                        INNER JOIN information_schema.key_column_usage AS pkey ON
+                            rcon.unique_constraint_catalog = pkey.constraint_catalog AND
+                            rcon.unique_constraint_schema = pkey.constraint_schema AND
+                            rcon.unique_constraint_name = pkey.constraint_name
+                        INNER JOIN information_schema.key_column_usage AS fkey ON
+                            rcon.constraint_catalog = fkey.constraint_catalog AND
+                            rcon.constraint_schema = fkey.constraint_schema AND
+                            rcon.constraint_name = fkey.constraint_name
+            );
+        """
+        await self.conn.raw_execute(query)
+
         # query table names in dependency order
         query = """
             WITH RECURSIVE
-                key_reference AS (
-                    SELECT
-                        rcon.unique_constraint_catalog AS primary_constraint_catalog,
-                        rcon.unique_constraint_schema AS primary_constraint_schema,
-                        rcon.unique_constraint_name AS primary_constraint_name,
-                        pkey.table_catalog AS primary_table_catalog,
-                        pkey.table_schema AS primary_table_schema,
-                        pkey.table_name AS primary_table_name,
-                        rcon.constraint_catalog AS foreign_constraint_catalog,
-                        rcon.constraint_schema AS foreign_constraint_schema,
-                        rcon.constraint_name AS foreign_constraint_name,
-                        fkey.table_catalog AS foreign_table_catalog,
-                        fkey.table_schema AS foreign_table_schema,
-                        fkey.table_name AS foreign_table_name
-                    
-                    FROM
-                        information_schema.referential_constraints AS rcon
-                            INNER JOIN information_schema.key_column_usage AS pkey ON
-                                rcon.unique_constraint_catalog = pkey.constraint_catalog AND
-                                rcon.unique_constraint_schema = pkey.constraint_schema AND
-                                rcon.unique_constraint_name = pkey.constraint_name
-                            INNER JOIN information_schema.key_column_usage AS fkey ON
-                                rcon.constraint_catalog = fkey.constraint_catalog AND
-                                rcon.constraint_schema = fkey.constraint_schema AND
-                                rcon.constraint_name = fkey.constraint_name
-                ),
                 dependencies(
                         depth,
                         parent_catalog,
@@ -119,7 +129,7 @@ class _CatalogSchemaBuilder:
                         child_schema,
                         child_name
                 ) AS (
-                    -- tables that have no foreign keys
+                    -- tables that have no foreign keys to other tables
                     SELECT
                         1 AS depth,
                         NULL::information_schema.sql_identifier,
@@ -130,6 +140,7 @@ class _CatalogSchemaBuilder:
                         tab.table_name
                     FROM (
                         (
+                            -- all tables (but not views)
                             SELECT
                                 table_catalog, table_schema, table_name
                             FROM
@@ -141,10 +152,16 @@ class _CatalogSchemaBuilder:
                         )
                         EXCEPT
                         (
+                            -- tables with foreign keys (excluding self-references)
                             SELECT
                                 foreign_table_catalog, foreign_table_schema, foreign_table_name
                             FROM
                                 key_reference
+                            WHERE
+                                -- exclude self-references (such as "users" table referencing "users" table)
+                                primary_table_catalog != foreign_table_catalog OR
+                                primary_table_schema != foreign_table_schema OR
+                                primary_table_name != foreign_table_name
                         )
                     ) AS tab
                 UNION
@@ -244,7 +261,7 @@ class _CatalogSchemaBuilder:
                     f"unrecognized database column type {column_type} in table {db_table}"
                 )
 
-            if column["is_nullable"] and column["column_default"] is not None:
+            if column["is_nullable"] and column["column_default"] is None:
                 outer_type = Optional[value_type]
             else:
                 outer_type = value_type
@@ -310,28 +327,20 @@ class _CatalogSchemaBuilder:
 
         query = """
             SELECT
-                fkey.constraint_name AS foreign_key_name,
-                fkey.table_schema AS foreign_key_schema,
-                fkey.table_name AS foreign_key_table,
-                fkey.column_name AS foreign_key_column,
-                pkey.constraint_name AS primary_key_name,
-                pkey.table_schema AS primary_key_schema,
-                pkey.table_name AS primary_key_table,
-                pkey.column_name AS primary_key_column
+                foreign_constraint_name AS foreign_key_name,
+                foreign_table_schema AS foreign_key_schema,
+                foreign_table_name AS foreign_key_table,
+                foreign_column_name AS foreign_key_column,
+                primary_constraint_name AS primary_key_name,
+                primary_table_schema AS primary_key_schema,
+                primary_table_name AS primary_key_table,
+                primary_column_name AS primary_key_column
             FROM
-                information_schema.referential_constraints ref_con
-                    INNER JOIN information_schema.key_column_usage pkey ON
-                        ref_con.unique_constraint_catalog = pkey.constraint_catalog AND
-                        ref_con.unique_constraint_schema = pkey.constraint_schema AND
-                        ref_con.unique_constraint_name = pkey.constraint_name
-                    INNER JOIN information_schema.key_column_usage fkey ON
-                        ref_con.constraint_catalog = fkey.constraint_catalog AND
-                        ref_con.constraint_schema = fkey.constraint_schema AND
-                        ref_con.constraint_name = fkey.constraint_name
+                key_reference
             WHERE
-                fkey.table_catalog = CURRENT_CATALOG
-                    AND fkey.table_schema = $1
-                    AND fkey.table_name = $2
+                foreign_table_catalog = CURRENT_CATALOG
+                    AND foreign_table_schema = $1
+                    AND foreign_table_name = $2
         """
         constraints = await self.conn.typed_fetch(
             _ReferenceConstraint, query, self.db_schema, table_schema.name
@@ -363,7 +372,7 @@ async def get_catalog_schema(conn: DatabaseClient, db_schema: str) -> CatalogSch
 
 
 def column_to_field(
-    column: ColumnSchema,
+    column: ColumnSchema, optional_default: bool = True
 ) -> Tuple[str, type, Field]:
     if keyword.iskeyword(column.name):
         field_name = f"{column.name}_"  # PEP 8: single trailing underscore to avoid conflicts with Python keyword
@@ -379,7 +388,7 @@ def column_to_field(
     default = MISSING
     if column.default is not None:
         default = column.default
-    elif is_optional_type(column.data_type):
+    elif optional_default and is_optional_type(column.data_type):
         default = None
 
     return (
@@ -389,10 +398,17 @@ def column_to_field(
     )
 
 
-def table_to_dataclass(table: TableSchema) -> DataClass:
-    "Generates a dataclass type corresponding to a table schema."
+def table_to_dataclass(table: TableSchema, optional_default: bool = True) -> DataClass:
+    """
+    Generates a dataclass type corresponding to a table schema.
 
-    fields = [column_to_field(column) for column in table.columns.values()]
+    :param table: The database table from which to produce a dataclass.
+    :param optional_default: Whether to assign a default value of `None` to fields with type `Optional[T]`.
+    """
+
+    fields = [
+        column_to_field(column, optional_default) for column in table.columns.values()
+    ]
     if keyword.iskeyword(table.name):
         class_name = f"{table.name}_"  # PEP 8: single trailing underscore to avoid conflicts with Python keyword
     else:
@@ -430,12 +446,11 @@ def catalog_to_dataclasses(catalog: CatalogSchema) -> List[DataClass]:
     return [table_to_dataclass(table) for table in catalog.tables.values()]
 
 
-def dataclasses_to_stream(types: List[DataClass], target: TextIO):
-    "Generates Python code corresponding to a dataclass type."
-
+def _header_to_stream(target: TextIO) -> None:
     print("# This source file has been generated by a tool, do not edit", file=target)
+    print("import enum", file=target)
     print("from dataclasses import dataclass, field", file=target)
-    print("from datetime import date, datetime, time", file=target)
+    print("from datetime import date, datetime, time, timedelta", file=target)
     print("from decimal import Decimal", file=target)
     print("from typing import Optional", file=target)
     print("from uuid import UUID", file=target)
@@ -444,39 +459,73 @@ def dataclasses_to_stream(types: List[DataClass], target: TextIO):
     print("from strong_typing.auxiliary import *", file=target)
     print(file=target)
 
+
+def dataclass_to_stream(typ: DataClass, target: TextIO) -> None:
+    "Generates Python code corresponding to a dataclass type."
+
+    print(file=target)
+    print("@dataclass", file=target)
+    print(f"class {typ.__name__}:", file=target)
+
+    # check if class has a doc-string other than the auto-generated string assigned by @dataclass
+    if typ.__doc__ and not re.match(
+        f"^{re.escape(typ.__name__)}[(].*[)]$", typ.__doc__
+    ):
+        if "\n" in typ.__doc__:
+            print('    """', file=target)
+            for line in textwrap.dedent(typ.__doc__).lstrip().splitlines():
+                if not line:
+                    # keep empty lines
+                    print(file=target)
+                else:
+                    # wrap long lines
+                    for wrapline in textwrap.wrap(
+                        line,
+                        width=119,
+                        initial_indent="    ",
+                        subsequent_indent="    ",
+                    ):
+                        print(wrapline, file=target)
+            print('    """', file=target)
+        else:
+            print(f"    {repr(typ.__doc__)}", file=target)
+        print(file=target)
+
+    # class variables (e.g. "primary_key")
+    field_names = [field.name for field in dataclasses.fields(typ)]
+    variables = {
+        name: value
+        for name, value in inspect.getmembers(typ, lambda m: not inspect.isroutine(m))
+        if not re.match(r"^__.+__$", name) and name not in field_names
+    }
+    if variables:
+        for name, value in variables.items():
+            print(f"    {name} = {repr(value)}", file=target)
+        print(file=target)
+
+    # table columns
+    for field in dataclasses.fields(typ):
+        type_name = python_type_to_str(field.type)
+        metadata = dict(field.metadata)
+        metadata.pop("description", None)
+        if field.default is not MISSING and metadata:
+            initializer = f" = field(default = {repr(field.default)}, metadata = {repr(metadata)})"
+        elif metadata:
+            initializer = f" = field(metadata = {repr(metadata)})"
+        elif field.default is not MISSING:
+            initializer = f" = {repr(field.default)}"
+        else:
+            initializer = ""
+        print(f"    {field.name}: {type_name}{initializer}", file=target)
+    print(file=target)
+
+
+def dataclasses_to_stream(types: List[DataClass], target: TextIO) -> None:
+    "Generates Python code corresponding to a set of dataclass types."
+
+    _header_to_stream(target)
     for typ in types:
-        print(file=target)
-        print("@dataclass", file=target)
-        print(f"class {typ.__name__}:", file=target)
-        if typ.__doc__:
-            if "\n" in typ.__doc__:
-                print('    """', file=target)
-                target.write(textwrap.indent(typ.__doc__, "    "))
-                print('    """', file=target)
-            else:
-                print(f"    {repr(typ.__doc__)}", file=target)
-            print(file=target)
-
-        # primary key
-        if getattr(typ, "primary_key", None) is not None:
-            print(f"    primary_key = {repr(typ.primary_key)}", file=target)
-            print(file=target)
-
-        # table columns
-        for field in dataclasses.fields(typ):
-            type_name = python_type_to_str(field.type)
-            metadata = dict(field.metadata)
-            metadata.pop("description", None)
-            if field.default is not MISSING and metadata:
-                initializer = f" = field(default = {repr(field.default)}, metadata = {repr(metadata)})"
-            elif metadata:
-                initializer = f" = field(metadata = {repr(metadata)})"
-            elif field.default is not MISSING:
-                initializer = f" = {repr(field.default)}"
-            else:
-                initializer = ""
-            print(f"    {field.name}: {type_name}{initializer}", file=target)
-        print(file=target)
+        dataclass_to_stream(typ, target)
 
 
 def dataclasses_to_code(types: List[DataClass]) -> str:
@@ -485,6 +534,94 @@ def dataclasses_to_code(types: List[DataClass]) -> str:
     return f.getvalue()
 
 
+@dataclasses.dataclass
+class EnumField:
+    """
+    A member in an enumeration class.
+
+    :param name: The name of the enumeration member.
+    :param value: The value of the enumeration member.
+    :param description: The doc-string associated with the member.
+    """
+
+    name: str
+    value: Any
+    description: Optional[str] = None
+
+
+def enum_class_to_stream(enum_class: enum.Enum, target: TextIO) -> None:
+    "Writes an enumeration class as a class definition."
+
+    print("@enum.unique", file=target)
+    print(f"class {enum_class.__name__}(enum.Enum):", file=target)
+    if enum_class.__doc__:
+        print(f"    {repr(enum_class.__doc__)}", file=target)
+        print(file=target)
+
+    for e in enum_class:
+        value = repr(e.value)
+        print(f"    {e.name} = {value}", file=target)
+
+    for e in enum_class:
+        if not e.__doc__ or e.__doc__ == enum_class.__doc__:
+            continue
+
+        print(
+            f"{enum_class.__name__}.{e.name}.__doc__ = {repr(e.__doc__)}",
+            file=target,
+        )
+
+    print(file=target)
+
+
+def enum_class_to_code(enum_class: enum.Enum) -> str:
+    "Returns an enumeration class as a class definition string."
+
+    with StringIO() as out:
+        enum_class_to_stream(enum_class, out)
+        return out.getvalue()
+
+
+def enum_to_class(
+    class_name: str, fields: List[EnumField], doc_string: str = None
+) -> type:
+    "Generates a class type corresponding to an enumeration."
+
+    enum_class = enum.Enum(
+        class_name,
+        {field.name: field.value for field in fields},
+        module=__name__,
+    )
+
+    # assign doc-string to class
+    if doc_string:
+        enum_class.__doc__ = doc_string
+
+    # assign doc-string to enum members
+    field_doc_string = {
+        field.name: field.description for field in fields if field.description
+    }
+    for e in enum_class:
+        description = field_doc_string.get(e.name)
+        if description:
+            e.__doc__ = description
+
+    return enum_class
+
+
+def classes_to_stream(types: List[DataClass], target: TextIO) -> None:
+    "Generates Python code corresponding to a set of enum class and dataclass types."
+
+    _header_to_stream(target)
+    for typ in types:
+        if is_type_enum(typ):
+            enum_class_to_stream(typ, target)
+        elif is_dataclass_type(typ):
+            dataclass_to_stream(typ, target)
+        else:
+            raise NotImplementedError(f"unsupported type: {typ}")
+
+
 async def main(output_path: str, db_schema: str) -> None:
     async with async_database.connection(ConnectionParameters()) as conn:
         catalog = await get_catalog_schema(conn, db_schema)
